from __future__ import annotations

from braket.ahs.atom_arrangement import SiteType
from braket.ahs.driving_field import DrivingField
from braket.ahs.shifting_field import ShiftingField
from braket.ahs.pattern import Pattern
from braket.ahs.field import Field
import braket.ir.ahs as braket_ir

from braket.ahs.analog_hamiltonian_simulation import AnalogHamiltonianSimulation
from braket.ahs.atom_arrangement import AtomArrangement

from typing import NoReturn, Optional, Tuple, Union

# import simplejson as json
from braket.ir.ahs import Program

import json
import numpy as np

import quera_ahs_utils.drive as drive
from quera_ahs_utils.quera_ir.task_specification import (QuEraTaskSpecification, 
    Lattice, EffectiveHamiltonian, RydbergHamiltonian, RabiFrequencyAmplitude,
    RabiFrequencyPhase, Detuning, GlobalField, LocalField)



__all__ = [
    'to_json_file',
    'from_json_file',
    'quera_json_to_ahs',
    'braket_sdk_to_quera_json',
]

def to_json_file(js:dict,json_filename:str,**json_options) -> NoReturn:
    """prints out a dictionary to a json file. 

    Args:
        js (dict): data to be serialaized.
        json_filename (str): filename to output to.
        **json_options: options that get passed to the json serializer as `json.dump(...,**json_options)`
        
    """
    with open(json_filename,"w") as IO:
        json.dump(js,IO,**json_options)


def from_json_file(json_filename:str) -> dict:
    """deserialize a json file. 

    Args:
        json_filename (str): the json file to deserialize. 

    Returns:
        dict: the json file deserialized as a python dict. 
    """
    with open(json_filename,"w") as IO:
        js = json.load(IO)

    return js


class quera_to_braket:

    @staticmethod
    def get_register(lattice: Lattice):
        register = AtomArrangement()
        for coord,filling in zip(lattice.sites, lattice.filling):
            site_type = SiteType.FILLED if filling==1 else SiteType.VACANT
            register.add(coord, site_type)
        
        return register

    @staticmethod
    def get_braket_field(quera_field: Optional[Union[GlobalField, LocalField]]):
        if isinstance(quera_field, GlobalField):
            pattern = Pattern("uniform")
            time_series = drive.time_series(quera_field.times, quera_field.values)
        elif isinstance(quera_field, LocalField):
            pattern = Pattern(quera_field.lattice_site_coefficients)
            time_series = drive.time_series(quera_field.times, quera_field.values)
        elif quera_field == None:
            return None
        else:
            raise TypeError("expecting quera_field to be one of quera_ir.GlobalField, quera_ir.LocalField.")
        
        return Field(time_series=time_series, pattern=pattern)    

    @staticmethod
    def get_amplitude(amplitude: RabiFrequencyAmplitude):
        return quera_to_braket.get_braket_field(amplitude.global_)

    @staticmethod
    def get_phase(phase: RabiFrequencyPhase):
        return quera_to_braket.get_braket_field(phase.global_)

    @staticmethod
    def get_detuning(detuning: Detuning):
        return quera_to_braket.get_braket_field(detuning.global_)

    @staticmethod
    def get_local_shifting_field(detuning: Detuning):
        return quera_to_braket.get_braket_field(detuning.local)
    
    @staticmethod
    def get_driving_field(rydberg: RydbergHamiltonian):
        return DrivingField(
            amplitude=quera_to_braket.get_amplitude(rydberg.rabi_frequency_amplitude), 
            detuning=quera_to_braket.get_detuning(rydberg.detuning), 
            phase=quera_to_braket.get_phase(rydberg.rabi_frequency_phase)
        )
    
    @staticmethod
    def get_shifting_field(rydberg: RydbergHamiltonian):
        magnitude = quera_to_braket.get_braket_field(rydberg.detuning.local)
        
        if magnitude != None:        
            return ShiftingField(magnitude)
        else:
            return None

    @staticmethod 
    def get_hamiltonian(effective_hamiltonian: EffectiveHamiltonian ):
        driving_field = quera_to_braket.get_driving_field(effective_hamiltonian.rydberg)
        shifting_field = quera_to_braket.get_shifting_field(effective_hamiltonian.rydberg)

        if shifting_field != None:
            return driving_field + shifting_field
        else:
            return driving_field
            
class braket_to_quera:

    @staticmethod
    def get_global_field(field: braket_ir.PhysicalField) -> GlobalField:
        if not isinstance(field, braket_ir.PhysicalField): 
            raise ValueError("expecting PhysicalField")
        
        if field.pattern != "uniform": 
            raise ValueError("global field must have 'uniform' pattern.")

        return GlobalField(
                times=field.time_series.times,
                values=field.time_series.values
            )
        
    @staticmethod
    def get_local_field(field: braket_ir.PhysicalField) -> LocalField:
        if not isinstance(field, braket_ir.PhysicalField): 
            raise ValueError("expecting PhysicalField")
        
        if field.pattern == "uniform": 
            raise ValueError("local field must have a list of real values for the pattern.")

        return LocalField(
                times=field.time_series.times,
                values=field.time_series.values,
                lattice_site_coefficients=field.pattern
            )
        
    @staticmethod
    def get_rabi_frequency_amplitude(driving: braket_ir.DrivingField):
        field = braket_to_quera.get_global_field(driving.amplitude)
        return RabiFrequencyAmplitude(
            global_=braket_to_quera.get_global_field(driving.amplitude)
        )
        
    @staticmethod
    def get_rabi_frequency_phase(driving: braket_ir.DrivingField):
        return RabiFrequencyPhase(
            global_=braket_to_quera.get_global_field(driving.phase)
        )
    
    @staticmethod
    def get_detuning(driving: braket_ir.DrivingField, shifting: Optional[braket_ir.ShiftingField]):
        if shifting == None:
            return Detuning(
                global_=braket_to_quera.get_global_field(driving.detuning)
            )
        else:
            return Detuning(
                global_=braket_to_quera.get_global_field(driving.detuning),
                local=braket_to_quera.get_local_field(shifting.magnitude)
            )

    @staticmethod
    def get_rydberg(driving: braket_ir.DrivingField, shifting: Optional[braket_ir.ShiftingField] = None):
        return RydbergHamiltonian(
            rabi_frequency_amplitude = braket_to_quera.get_rabi_frequency_amplitude(driving),
            rabi_frequency_phase=braket_to_quera.get_rabi_frequency_phase(driving),
            detuning=braket_to_quera.get_detuning(driving, shifting)
        )


    @staticmethod
    def get_effective_hamiltonian(hamiltonian_ir: braket_ir.Hamiltonian):
        
        driving_fields = hamiltonian_ir.drivingFields
        shifting_fields = hamiltonian_ir.shiftingFields
        
        if len(driving_fields) != 1: raise ValueError("QuEra IR only supports exactly one set of driving fields")
        if len(shifting_fields) > 1:  raise ValueError("QuEra IR only supports at most one set of shifting fields")
        
        if len(shifting_fields) == 0:
            return EffectiveHamiltonian(
                rydberg=braket_to_quera.get_rydberg(driving_fields[0])
            )
        else:
            return EffectiveHamiltonian(
                rydberg=braket_to_quera.get_rydberg(driving_fields[0], shifting_fields[0])
            )

    @staticmethod
    def get_lattice(setup: braket_ir.Setup):
        return Lattice(sites=setup.ahs_register.sites, filling=setup.ahs_register.filling)

def quera_task_to_braket_ahs(task_specification: QuEraTaskSpecification) -> Tuple[int,AnalogHamiltonianSimulation]:
    """Convert a QuEra compatible program to a braket AHS program. 

    Args:
        js (QuEraTaskSpecification): An object containing a program formatted to be accepted by 
        the QuEra API. 

    Returns:
        Tuple[int,AnalogHamiltonianSimulation]: A tuple continaing
            the number of shots as the first element and the ahs program 
            as the second argument. 
    """

    return task_specification.nshots, AnalogHamiltonianSimulation(
            register=quera_to_braket.get_register(task_specification.lattice), 
            hamiltonian=quera_to_braket.get_hamiltonian(task_specification.effective_hamiltonian)
        )

def braket_ahs_to_quera_task(ahs : AnalogHamiltonianSimulation, shots: int = 1) -> QuEraTaskSpecification:
    """Translates Braket AHS IR program to Quera-TaskSpecification object.

    Args:
        ahs (AnalogHamiltonianSimulation): AHS object from braket SDK
        shots (int): The number of shots to run this program

    Returns:
        QuEraTaskSpecification:  QuEraTaskSpecification object representation of program
    """

    ahs_ir = ahs.to_ir()
    # QuEra IR Program
    return QuEraTaskSpecification(
        nshots=shots,
        lattice=braket_to_quera.get_lattice(ahs_ir.setup),
        effective_hamiltonian=braket_to_quera.get_effective_hamiltonian(ahs_ir.hamiltonian)
    )



import numpy as np
import matplotlib.pyplot as plt
import networkx as nx
from braket.ahs.atom_arrangement import SiteType
from braket.ahs.driving_field import DrivingField
from braket.ahs.shifting_field import ShiftingField

from braket.tasks.analog_hamiltonian_simulation_quantum_task_result import AnalogHamiltonianSimulationQuantumTaskResult
from braket.ahs.atom_arrangement import AtomArrangement

__all__ = [
    "show_register",
    "show_global_drive",
    'show_local_shift',
    'show_drive_and_shift',
    'show_final_avg_density',
    'plot_avg_density',
]

def show_register(
    register: AtomArrangement, 
    blockade_radius: float=0.0, 
    what_to_draw: str="bond", 
    show_atom_index:bool=True
):
    """Plot the given register 

        Args:
            register (AtomArrangement): A given register
            blockade_radius (float): Default is 0. The blockade radius for the register.
            what_to_draw (str): Default is "bond". Either "bond" or "circle" to indicate the blockade region. 
            show_atom_index (bool): Default is True. Choose if each atom's index is displayed over the atom itself in the resulting figure. 
        
    """
    filled_sites = [site.coordinate for site in register if site.site_type == SiteType.FILLED]
    empty_sites = [site.coordinate for site in register if site.site_type == SiteType.VACANT]
    
    fig = plt.figure(figsize=(7, 7))
    if filled_sites:
        plt.plot(np.array(filled_sites)[:, 0], np.array(filled_sites)[:, 1], 'r.', ms=15, label='filled')
    if empty_sites:
        plt.plot(np.array(empty_sites)[:, 0], np.array(empty_sites)[:, 1], 'k.', ms=5, label='empty')
    plt.legend(bbox_to_anchor=(1.1, 1.05))
    
    if show_atom_index:
        for idx, site in enumerate(register):
            plt.text(*site.coordinate, f"  {idx}", fontsize=12)
    
    if blockade_radius > 0 and what_to_draw=="bond":
        for i in range(len(filled_sites)):
            for j in range(i+1, len(filled_sites)):            
                dist = np.linalg.norm(np.array(filled_sites[i]) - np.array(filled_sites[j]))
                if dist <= blockade_radius:
                    plt.plot([filled_sites[i][0], filled_sites[j][0]], [filled_sites[i][1], filled_sites[j][1]], 'b')
                    
    if blockade_radius > 0 and what_to_draw=="circle":
        for site in filled_sites:
            plt.gca().add_patch( plt.Circle((site[0],site[1]), blockade_radius/2, color="b", alpha=0.3) )
        plt.gca().set_aspect(1)
    plt.show()


def show_global_drive(drive, axes=None, **plot_ops):
    """Plot the driving field
        Args:
            drive (DrivingField): The driving field to be plot
            axes: Default is None. matplotlib axis to draw on
            **plot_ops: options passed to matplitlib.pyplot.plot
    """   

    data = {
        'amplitude [rad/s]': drive.amplitude.time_series,
        'detuning [rad/s]': drive.detuning.time_series,
        'phase [rad]': drive.phase.time_series,
    }


    if axes is None:
        fig, axes = plt.subplots(3, 1, figsize=(7, 7), sharex=True)

    for ax, data_name in zip(axes, data.keys()):
        if data_name == 'phase [rad]':
            ax.step(data[data_name].times(), data[data_name].values(), '.-', where='post',**plot_ops)
        else:
            ax.plot(data[data_name].times(), data[data_name].values(), '.-',**plot_ops)
        ax.set_ylabel(data_name)
        ax.grid(ls=':')
    axes[-1].set_xlabel('time [s]')
    plt.tight_layout()
    plt.show()


def show_local_shift(shift:ShiftingField):
    """Plot the shifting field
        Args:
            shift (ShiftingField): The shifting field to be plot
    """       
    data = shift.magnitude.time_series
    pattern = shift.magnitude.pattern.series
    
    plt.plot(data.times(), data.values(), '.-', label="pattern: " + str(pattern))
    plt.xlabel('time [s]')
    plt.ylabel('shift [rad/s]')
    plt.legend()
    plt.tight_layout()
    plt.show()

    
def show_drive_and_shift(drive: DrivingField, shift: ShiftingField):
    """Plot the driving and shifting fields
    
        Args:
            drive (DrivingField): The driving field to be plot
            shift (ShiftingField): The shifting field to be plot
    """        
    drive_data = {
        'amplitude [rad/s]': drive.amplitude.time_series,
        'detuning [rad/s]': drive.detuning.time_series,
        'phase [rad]': drive.phase.time_series,
    }
    
    fig, axes = plt.subplots(4, 1, figsize=(7, 7), sharex=True)
    for ax, data_name in zip(axes, drive_data.keys()):
        if data_name == 'phase [rad]':
            ax.step(drive_data[data_name].times(), drive_data[data_name].values(), '.-', where='post')
        else:
            ax.plot(drive_data[data_name].times(), drive_data[data_name].values(), '.-')
        ax.set_ylabel(data_name)
        ax.grid(ls=':')
        
    shift_data = shift.magnitude.time_series
    pattern = shift.magnitude.pattern.series   
    axes[-1].plot(shift_data.times(), shift_data.values(), '.-', label="pattern: " + str(pattern))
    axes[-1].set_ylabel('shift [rad/s]')
    axes[-1].set_xlabel('time [s]')
    axes[-1].legend()
    axes[-1].grid()
    plt.tight_layout()
    plt.show()


def show_final_avg_density(result: AnalogHamiltonianSimulationQuantumTaskResult):
    """Showing a bar plot for the average Rydberg densities from the result

        Args:
            result (AnalogHamiltonianSimulationQuantumTaskResult): The result 
                from which the aggregated state counts are obtained
    """    
    avg_density = get_avg_density(result)
    
    plt.bar(range(len(avg_density)), avg_density)
    plt.xlabel("Indices of atoms")
    plt.ylabel("Average Rydberg density")
    plt.show()


def plot_avg_density(densities, register, with_labels = True, custom_axes = None, cmap = None):
    """Plot the average Rydberg densities for any 2D geometry from the result

        Args:
            densities (List[Float]): The average Rydberg densities from the result, obtainable via
               invoking get_avg_density() on an AnalogHamiltonianSimulationQuantumTaskResult instance.

            register (AtomArrangement): The register used in creating the Hamiltonian.

            with_labels (Boolean): Default is True. Choose if each atom's index is displayed over the atom itself in the resulting figure. 
                Default is True.

            custom_axes (matplotlib.axes.Axes): Default is None. If argument is given, the plot will use the supplied
                axis for displaying data and the function will not return anything. Otherwise, a new matplotlib Figure and
                Axes will be generated and returned.

            cmap (matplotlib.colors.Colormap): Default is None. Defines the colormap that the plot uses to map the average density values
                to the colors of each plotted atom. When Default value is used a the resulting plot uses a Colormap that is given by 
                `matplotlib.pyplot.cm.bwr` which is gradient from red to blue with white in the middle.  
                
        Returns:
            Tuple[Optional[matplotlib.figure.Figure],matplotlib.axes.Axes]]: returns the Figure and the Axes object used to create the plot if `custom_axes`
                is not given, otherwise the function returns None
    """
    
    # get atom coordinates
    atom_coords = list(zip(register.coordinate_list(0), register.coordinate_list(1)))
    # convert all to micrometers
    atom_coords = [(atom_coord[0] * 10**6, atom_coord[1] * 10**6) for atom_coord in atom_coords]

    pos = {i:coord for i,coord in enumerate(atom_coords)}
           
    # get colors
    vmin = 0
    vmax = 1
    if cmap is None:
        cmap = plt.cm.bwr
    
    # construct graph
    g = nx.Graph()
    g.add_nodes_from(list(range(len(densities))))
    
    # construct plot
    if custom_axes is None:
        return_fig = True
        fig, ax = plt.subplots()
    else:
        ax = custom_axes
        return_fig = False
    
    nx.draw(g, 
            pos,
            node_color=densities,
            cmap=cmap,
            node_shape="o",
            vmin=vmin,
            vmax=vmax,
            font_size=9,
            with_labels=with_labels,
            ax = custom_axes if custom_axes is not None else ax)
        
    ## Set axes
    ax.set_axis_on()
    ax.tick_params(left=True, 
                   bottom=True, 
                   top=True,
                   right=True,
                   labelleft=True, 
                   labelbottom=True, 
                   # labeltop=True,
                   # labelright=True,
                   direction="in")
    ## Set colorbar
    sm = plt.cm.ScalarMappable(cmap=cmap, norm=plt.Normalize(vmin=vmin, vmax=vmax))
    sm.set_array([])

    
    ax.ticklabel_format(style="sci", useOffset=False)
    
    # set titles on x and y axes
    plt.xlabel("x [μm]")
    plt.ylabel("y [μm]")
    
    

    cbar_label = "Rydberg Density"
        
    plt.colorbar(sm, ax=ax, label=cbar_label)
    
    if return_fig:
        return fig,ax
    else:
        return None,ax




import json
import unittest
import sys
import os

from braket.ahs.atom_arrangement import SiteType
from braket.timings.time_series import TimeSeries
from braket.ahs.driving_field import DrivingField

from braket.ahs.analog_hamiltonian_simulation import AnalogHamiltonianSimulation
from braket.ahs.atom_arrangement import AtomArrangement,SiteType

from decimal import Decimal
from quera_ahs_utils.ir import braket_ahs_to_quera_task, quera_task_to_braket_ahs
from quera_ahs_utils.quera_ir.task_specification import QuEraTaskSpecification, discretize_list
import numpy as np



class IrModule(unittest.TestCase):

    def pulse_values(self):
        # Define a set of time points
        time_points = [0, 0.6e-6, 3.4e-6, 4e-6]

        # Define the strength of the transverse field Ω
        amplitude_min = 0
        amplitude_max = 10e6  # rad / sec

        # Define the strength of the detuning Δ
        Delta_initial = -20e6     # rad / sec
        Delta_final = 20e6 

        amplitude_values = [amplitude_min, amplitude_max, amplitude_max, amplitude_min]  # piecewise linear
        detuning_values = [Delta_initial, Delta_initial, Delta_final, Delta_final]  # piecewise linear
        phase_values = [0, 0, 0, 0]  # piecewise constant

        return time_points,amplitude_values,detuning_values,phase_values


    def generate_ahs_program(self):
        nshots = 100
        register = AtomArrangement()
        register.add((0.0,0.0),SiteType.FILLED)
        register.add((0.0,5.0e-6),SiteType.FILLED)
        register.add((0.0,1e-5),SiteType.VACANT)

        time_points,amplitude_values,detuning_values,phase_values = self.pulse_values()

        amplitude = TimeSeries()
        for t,v in zip(time_points,amplitude_values):
            amplitude.put(t,v)

        detuning = TimeSeries()
        for t,v in zip(time_points,detuning_values):
            detuning.put(t,v)

        phase = TimeSeries()
        for t,v in zip(time_points,phase_values):
            phase.put(t,v)

        drive = DrivingField(
            amplitude=amplitude, 
            detuning=detuning, 
            phase=phase
        )

        return nshots,AnalogHamiltonianSimulation(
            register=register, 
            hamiltonian=drive
        )

    def generate_quera_ir(self):
        nshots = 100

        sites = [
            [0.0,0.0],
            [0.0,5.0e-6],
            [0.0,1.0e-5],
        ]
        filling = [1,1,0]

        time_points,amplitude_values,detuning_values,phase_values = self.pulse_values()

        # QuEra IR Program
        quera_ir = dict()
        quera_ir["nshots"] = nshots
        quera_ir["lattice"] = {"sites":sites,"filling":filling}
        quera_ir["effective_hamiltonian"] = {
            "rydberg": {
                "rabi_frequency_amplitude": {"global": {"times": time_points, "values": amplitude_values}},
                "rabi_frequency_phase": {"global": {"times": time_points, "values":phase_values}},
                "detuning": {
                    "global": {"times": time_points, "values": detuning_values},
                },
            }
        }

        return quera_ir

    def test_quera_to_braket(self):
        nshots,ahs_program = self.generate_ahs_program()
        quera_ir = self.generate_quera_ir()
        translated_quera_ir = braket_ahs_to_quera_task(ahs_program,shots=nshots)
        self.assertEqual(translated_quera_ir.json(by_alias=True, exclude_unset=True),json.dumps(quera_ir))

    def test_braket_to_quera(self):
        nshots,ahs_program = self.generate_ahs_program()
        quera_ir = QuEraTaskSpecification(**self.generate_quera_ir())
        translated_nshots,translated_ahs_program = quera_task_to_braket_ahs(quera_ir)
        self.assertEqual(ahs_program.to_ir().json(),translated_ahs_program.to_ir().json())
        self.assertEqual(nshots,translated_nshots)


    def test_discretize(self):
        my_list = [0.0134234,0.324,10.000001]
        my_discrete_list = discretize_list(my_list, 0.001)
        self.assertEqual(my_discrete_list, [Decimal('0.013'),Decimal('0.324'),Decimal('10.0')])



import numpy as np
from braket.tasks.analog_hamiltonian_simulation_quantum_task_result import AnalogHamiltonianSimulationQuantumTaskResult
from typing import Dict

__all__ = [
    'get_counts',
    'get_avg_density',
]


def get_counts(result: AnalogHamiltonianSimulationQuantumTaskResult) -> Dict[str, int]:
    """Aggregate state counts from AHS shot results

        Args:
            result (AnalogHamiltonianSimulationQuantumTaskResult): The result 
                from which the aggregated state counts are obtained

        Returns:
            Dict[str, int]: number of times each state configuration is measured

        Notes: We use the following convention to denote the state of an atom (site):
            e: empty site
            r: Rydberg state atom
            g: ground state atom
    """

    state_counts = Counter()
    states = ['e', 'r', 'g']
    for shot in result.measurements:
        pre = shot.pre_sequence
        post = shot.post_sequence
        state_idx = np.array(pre) * (1 + np.array(post))
        state = "".join(map(lambda s_idx: states[s_idx], state_idx))
        state_counts.update((state,))

    return dict(state_counts)

def get_avg_density(result: AnalogHamiltonianSimulationQuantumTaskResult) -> np.ndarray:
    """Get the average Rydberg densities from the result

        Args:
            result (AnalogHamiltonianSimulationQuantumTaskResult): The result 
                from which the aggregated state counts are obtained

        Returns: 
            ndarray: The average densities from the result
    """    

    measurements = result.measurements
    postSeqs = [measurement.post_sequence for measurement in measurements]
    postSeqs = 1 - np.array(postSeqs) # change the notation such 1 for rydberg state, and 0 for ground state
    
    avg_density = np.sum(postSeqs, axis=0)/len(postSeqs)
    
    return avg_density





from pydantic import BaseModel

__all__ = ["QuEraCapabilities"]


class RydbergGlobalCapabilities(BaseModel):
    rabi_frequency_min: float
    rabi_frequency_max: float
    rabi_frequency_resolution: float
    rabi_frequency_slew_rate_max: float
    detuning_min: float
    detuning_max: float
    detuning_resolution: float
    detuning_slew_rate_max: float
    phase_min: float
    phase_max: float
    phase_resolution: float
    time_min: float
    time_max: float
    time_resolution: float
    time_delta_min: float

class RydbergLocalCapabilities(BaseModel):
    detuning_min: float
    detuning_max: float
    detuning_slew_rate_max: float
    site_coefficient_min: float
    site_coefficient_max: float
    number_local_detuning_sites: int
    spacing_radial_min: float
    time_resolution: float
    time_delta_min: float

class RydbergCapabilities(BaseModel):
    c6_coefficient: float
    global_: RydbergGlobalCapabilities
    local: RydbergLocalCapabilities

    class Config:
        allow_population_by_field_name = True
        fields = {
            'global_': 'global'
        }

class LatticeGeometryCapabilities(BaseModel):
    spacing_radial_min: float
    spacing_vertical_min: float
    position_resolution: float
    number_sites_max: int
    
class LatticeAreaCapabilities(BaseModel):
    width: float
    height: float

class LatticeCapabilities(BaseModel):
    number_qubits_max: int
    area: LatticeAreaCapabilities
    geometry: LatticeGeometryCapabilities

class TaskCapabilities(BaseModel):
    number_shots_min: int
    number_shots_max: int

class DeviceCapabilities(BaseModel):
    task: TaskCapabilities
    lattice: LatticeCapabilities
    rydberg: RydbergCapabilities

class QuEraCapabilities(BaseModel):
    version: str
    capabilities: DeviceCapabilities


from pydantic import BaseModel
from typing import Optional, List, Tuple, Union
from decimal import Decimal

from quera_ahs_utils.quera_ir.capabilities import QuEraCapabilities


__all__ = [
    "QuEraTaskSpecification"
]

# TODO: add version to these models.

FloatType = Union[Decimal, float]

def discretize_list(list_of_values: list, resolution: FloatType):
    resolution = Decimal(str(float(resolution)))
    return [Decimal(value).quantize(resolution) for value in list_of_values]

class GlobalField(BaseModel):
    times: List[FloatType]
    values: List[FloatType]
    
    def __hash__(self):
        return hash((GlobalField, tuple(self.times), tuple(self.values)))
        
class LocalField(BaseModel):
    times: List[FloatType]
    values: List[FloatType]
    lattice_site_coefficients: List[FloatType]
    
    def __hash__(self):
        return hash((LocalField, tuple(self.times), tuple(self.values), tuple(self.lattice_site_coefficients)))

class RabiFrequencyAmplitude(BaseModel):
    global_: GlobalField
    
    class Config:
        allow_population_by_field_name = True
        fields = {
            'global_': 'global'
        }
    
    def __hash__(self):
        return hash((RabiFrequencyAmplitude, self.global_))
    
    def discretize(self, task_capabilities: QuEraCapabilities):
        global_time_resolution = task_capabilities.capabilities.rydberg.global_.time_resolution
        global_value_resolution =  task_capabilities.capabilities.rydberg.global_.rabi_frequency_resolution
        
        return RabiFrequencyAmplitude(
            global_ = GlobalField(
                times = discretize_list(self.global_.times, global_time_resolution),
                values = discretize_list(self.global_.values, global_value_resolution)
            ) 
        )

class RabiFrequencyPhase(BaseModel):
    global_: GlobalField
    
    class Config:
        allow_population_by_field_name = True
        fields = {
            'global_': 'global'
        }
        
    def __hash__(self):
        return hash((RabiFrequencyPhase, self.global_))
    
    def discretize(self, task_capabilities: QuEraCapabilities):
        global_time_resolution = task_capabilities.capabilities.rydberg.global_.time_resolution
        global_value_resolution =  task_capabilities.capabilities.rydberg.global_.phase_resolution
        
        return RabiFrequencyPhase(global_=GlobalField(
                    times = discretize_list(self.global_.times, global_time_resolution),
                    values = discretize_list(self.global_.values, global_value_resolution)
                )
            )
    
class Detuning(BaseModel):
    global_: GlobalField
    local: Optional[LocalField]
    
    class Config:
        allow_population_by_field_name = True
        fields = {
            'global_': 'global'
        }

    def __hash__(self):
        return hash((Detuning, self.global_, self.local))
    
    def discretize(self, task_capabilities: QuEraCapabilities):
        global_time_resolution = task_capabilities.capabilities.rydberg.global_.time_resolution
        global_value_resolution =  task_capabilities.capabilities.rydberg.global_.detuning_resolution
        local_time_resolution = task_capabilities.capabilities.rydberg.local.time_resolution

        if self.local != None:
            self.local = LocalField(
                    times = discretize_list(self.local.times, local_time_resolution), 
                    values = self.local.values,
                    lattice_site_coefficients=self.local.lattice_site_coefficients
                )


        return Detuning(
                global_ = GlobalField(
                    times = discretize_list(self.global_.times, global_time_resolution),
                    values = discretize_list(self.global_.values, global_value_resolution)
                ),
                local = self.local
            )
    
class RydbergHamiltonian(BaseModel):
    rabi_frequency_amplitude: RabiFrequencyAmplitude
    rabi_frequency_phase: RabiFrequencyPhase
    detuning: Detuning
    
    def __hash__(self):
        return hash((RydbergHamiltonian, self.rabi_frequency_amplitude, self.rabi_frequency_phase, self.detuning))
    
    def discretize(self, task_capabilities: QuEraCapabilities):
        return RydbergHamiltonian(
            rabi_frequency_amplitude = self.rabi_frequency_amplitude.discretize(task_capabilities),
            rabi_frequency_phase = self.rabi_frequency_phase.discretize(task_capabilities),
            detuning = self.detuning.discretize(task_capabilities)
        )

    
class EffectiveHamiltonian(BaseModel):
    rydberg: RydbergHamiltonian
    
    def __hash__(self):
        return hash((EffectiveHamiltonian, self.rydberg))
    
    def discretize(self, task_capabilities: QuEraCapabilities):
        return EffectiveHamiltonian(rydberg = self.rydberg.discretize(task_capabilities))

class Lattice(BaseModel):
    sites: List[Tuple[FloatType, FloatType]]
    filling: List[int]
    
    def __hash__(self):
        return hash((Lattice, tuple(self.sites), tuple(self.filling)))
    
    def discretize(self, task_capabilities: QuEraCapabilities):
        position_resolution = task_capabilities.capabilities.lattice.geometry.position_resolution
        return Lattice(
            sites = [discretize_list(site,position_resolution) for site in self.sites],
            filling = self.filling
        )
            
    
class QuEraTaskSpecification(BaseModel):
    nshots: int
    lattice: Lattice
    effective_hamiltonian: EffectiveHamiltonian
    
    def __hash__(self):
        return hash((QuEraTaskSpecification, self.nshots, self.lattice, self.effective_hamiltonian))
    
    def discretize(self, task_capabilities: QuEraCapabilities):
        return QuEraTaskSpecification(
            nshots = self.nshots,
            lattice = self.lattice.discretize(task_capabilities),
            effective_hamiltonian = self.effective_hamiltonian.discretize(task_capabilities)
        )
    



import unittest
from braket.timings.time_series import TimeSeries
from braket.ahs.driving_field import DrivingField

import quera_ahs_utils.drive as drive
import numpy as np

class DriveModule(unittest.TestCase):
    
    @staticmethod
    def time_series_all_close(first: TimeSeries, second: TimeSeries, reltol: float = 1.05e-8 ):
        if len(first) != len(second):
            return False
        if len(first) == 0:
            return True
        first_times = first.times()
        second_times = second.times()
        first_values = first.values()
        second_values = second.values()
        for index in range(len(first)):
            tolerance = max(abs(first_times[index]), abs(second_times[index])) * reltol
            if abs(first_times[index] - second_times[index]) >= tolerance:
                return False
            
            tolerance = max(abs(first_values[index]), abs(second_values[index])) * reltol
            if abs(first_values[index] - second_values[index]) >= tolerance:
                return False
            
        return True
    
    def test_time_series(self):
        times = [0.0, 0.1, 0.5, 1.0]
        values = [0.0, 3.0, 2.3, 1.0]
        
        test_time_series = TimeSeries()
        test_time_series.put(0.0, 0.0)
        test_time_series.put(0.1, 3.0)
        test_time_series.put(0.5, 2.3)
        test_time_series.put(1.0, 1.0)
        
        generated_time_series = drive.time_series(times, values)
        
        self.assertEqual(generated_time_series.times(),test_time_series.times())    
        self.assertEqual(generated_time_series.values(),test_time_series.values())
        
    def test_get_time_sample_value_1(self):
        times = [0.0, 0.5, 2.0, 3.0]
        values = [0.0, 1.0, 3.0, 1.0]
        
        time_series = drive.time_series(times, values)

        samples = [
            drive.get_time_series_value(time_series, -0.5),
            drive.get_time_series_value(time_series,  0.25),
            drive.get_time_series_value(time_series,  2.0),
            drive.get_time_series_value(time_series, 4.0)
        ]
        self.assertEqual(samples,[0.0,0.5,3.0,1.0])
        
    def test_get_time_sample_value_2(self):
        times = [0.0, 0.5, 2.0, 3.0]
        values = [0.0, 1.0, 3.0, 1.0]
        
        time_series = drive.time_series(times, values)

        samples = [
            drive.get_time_series_value(time_series, -0.5, piecewise_constant=True),
            drive.get_time_series_value(time_series,  0.25, piecewise_constant=True),
            drive.get_time_series_value(time_series,  2.0, piecewise_constant=True),
            drive.get_time_series_value(time_series, 4.0, piecewise_constant=True)
        ]
        self.assertEqual(samples,[0.0,0.0,3.0,1.0])
        
    def test_slice_time_series_1(self):
        original_time_series = drive.time_series(
            [0.0, 0.5, 1.5, 2.0],
            [0.0, 1.0, 2.5, 0.0]
        )
        
        sliced_time_series = drive.slice_time_series(original_time_series,0.25,1)
        test_time_series = drive.time_series(
            [0.0,0.25,0.75],
            [0.5,1.0,1.75]
        )
        self.assertEqual(sliced_time_series.times(),test_time_series.times())    
        self.assertEqual(sliced_time_series.values(),test_time_series.values())
        
        
    def test_slice_time_series_2(self):
        original_time_series = drive.time_series(
            [0.0, 0.5, 1.5, 2.0],
            [0.0, 1.0, 2.5, 0.0]
        )
        
        sliced_time_series = drive.slice_time_series(original_time_series,0.0,2.0)
        self.assertEqual(sliced_time_series.times(),original_time_series.times())    
        self.assertEqual(sliced_time_series.values(),original_time_series.values())
        
    def test_slice_time_series_3(self):
        original_time_series = drive.time_series(
            [0.0, 0.5, 1.5, 2.0],
            [0.0, 1.0, 2.5, 0.0]
        )
        
        sliced_time_series = drive.slice_time_series(original_time_series,0.0,1.0)
        test_time_series = drive.time_series(
            [0.0,0.5,1.0],
            [0.0,1.0,1.75]
        )
        
        self.assertEqual(sliced_time_series.times(),test_time_series.times())    
        self.assertEqual(sliced_time_series.values(),test_time_series.values())
    

        


from enum import Enum
from pydantic import BaseModel, conlist, conint
from typing import Callable, Optional, List, Tuple
import numpy as np

__all__ = [
    "QuEraTaskResults",  
    "TaskProbabilities"  
]

# TODO: add version to these models.

class QuEraShotStatusCode(str, Enum):
    Completed = "Completed"
    MissingPreSequence = "MissingPreSequence"
    MissingPostSequence = "MissingPostSequence"
    MissingMeasurement = "MissingMeasurement"

class QuEraTaskStatusCode(str, Enum):
    Created = "Created"
    Running = "Running"
    Completed = "Completed"
    Failed = "Failed"
    Cancelled = "Cancelled"

class QuEraShotResult(BaseModel):
    shot_status: QuEraShotStatusCode = QuEraShotStatusCode.MissingMeasurement
    pre_sequence: conlist(conint(ge=0, le=1), min_items=0) = []
    post_sequence: conlist(conint(ge=0, le=1), min_items=0) = []

class TaskProbabilities(BaseModel):
    probabilities: List[Tuple[Tuple[str,str],float]]
    
    def simulate_task_results(self, shots = 1) -> 'QuEraTaskResults':
        bit_strings, probabilities = zip(*self.probabilities)
        
        indices = np.random.choice(
            len(probabilities), 
            p=probabilities,
            size=shots
        )
        shot_outputs = []
        for index in indices:
            pre_string, post_string = bit_strings[index]
            pre_sequence = [int(bit) for bit in pre_string]
            post_sequence = [int(bit) for bit in post_string]
            
            shot_outputs.append(
                QuEraShotResult(
                    shot_status = QuEraShotStatusCode.Completed,
                    pre_sequence = pre_sequence,
                    post_sequence = post_sequence
                )
            )
        
        return QuEraTaskResults(
            task_status=QuEraTaskStatusCode.Completed,
            shot_outputs=shot_outputs
        )

class QuEraTaskResults(BaseModel):
    task_status: QuEraTaskStatusCode = QuEraTaskStatusCode.Failed
    shot_outputs: conlist(QuEraShotResult, min_items=0) = []
    
    def export_as_probabilities(self) -> TaskProbabilities:
        """converts from shot results to probabilities

        Returns:
            TaskProbabilities: The task results as probabilties
        """
        counts = dict()
        nshots = len(self.shot_outputs)
        for shot_result in self.shot_outputs:
            
            pre_sequence_str = "".join(
                str(bit) for bit in shot_result.pre_sequence
            )
            
            post_sequence_str = "".join(
                str(bit) for bit in shot_result.post_sequence
            )
            
            configuration = (pre_sequence_str, post_sequence_str)
            # iterative average
            current_count = counts.get(configuration, 0)
            counts[configuration] = current_count + 1

        probabilities = [(config,count/nshots) for config,count in counts.items()]
        return TaskProbabilities(probabilities=probabilities)
    
    def post_process(self, keep_shot_result: Optional[Callable] = None, args = ()) -> 'QuEraTaskResults':
        
        if keep_shot_result == None:
            filter_func = \
                lambda shot_result: \
                    all(bit==1 for bit in shot_result.pre_sequence)
        else:
            filter_func = lambda shot_result: \
                keep_shot_result(shot_result, *args)
        
        return QuEraTaskResults(
            task_status=self.task_status,
            shot_outputs=list(
                filter(filter_func, self.shot_outputs)
            )
        )




import dataclasses
from itertools import product
import numpy as np
from typing import Tuple,Union, NoReturn, Optional
from decimal import Decimal

from braket.aws import AwsDevice
from braket.ahs.analog_hamiltonian_simulation import AnalogHamiltonianSimulation
from braket.ahs.driving_field import DrivingField
from braket.ahs.shifting_field import ShiftingField
from braket.ahs.hamiltonian import Hamiltonian
from braket.ahs.field import Field
from braket.task_result import AnalogHamiltonianSimulationTaskResult
from braket.ahs.atom_arrangement import AtomArrangement
from braket.ahs.pattern import Pattern



__all__ = [
    'generate_parallel_register',
    'parallelize_field',
    'parallelize_hamiltonian',
    'parallelize_ahs',
    'get_shots_braket_sdk_results',
    'parallelize_quera_json',
    'get_shots_quera_results',
]

def generate_parallel_register(
    register:AtomArrangement,
    qpu:AwsDevice,
    interproblem_distance:Union[float,Decimal]
) -> Tuple[AtomArrangement,dict]:
    """generate grid of parallel registers from a single register. 

    Args:
        register (AtomArrangement): The register set to parallelize
        qpu (AwsDevice): The QPU that will be running the parallel job
        interproblem_distance (UnionType[float,Decimal]): _description_

    Returns:
        _type_: _description_
    """
    if len(register) > 1:
        x_min = min(*[site.coordinate[0] for site in register])
        x_max = max(*[site.coordinate[0] for site in register])
        y_min = min(*[site.coordinate[1] for site in register])
        y_max = max(*[site.coordinate[1] for site in register])
    else:
        x_min = x_max = 0
        y_min = y_max = 0

    single_problem_width = x_max - x_min
    single_problem_height = y_max - y_min

    # get values from device capabilities
    field_of_view_width = qpu.properties.paradigm.lattice.area.width
    field_of_view_height = qpu.properties.paradigm.lattice.area.height
    n_site_max = qpu.properties.paradigm.lattice.geometry.numberSitesMax
    
    batch_mapping = dict()
    parallel_register = AtomArrangement()

    atom_number = 0 #counting number of atoms added
    
    ix = 0
    while True:
        x_shift = ix * (single_problem_width + interproblem_distance)
        # reached the maximum number of batches possible given n_site_max
        if atom_number + len(register) > n_site_max: break
        # reached the maximum number of batches possible along x-direction
        if x_shift + single_problem_width > field_of_view_width: break 
        
        iy = 0
        while True:
            y_shift = iy * (single_problem_height + interproblem_distance)
            # reached the maximum number of batches possible given n_site_max
            if atom_number + len(register) > n_site_max: break 
            # reached the maximum number of batches possible along y-direction
            if y_shift + single_problem_height > field_of_view_height: 
                ix += 1
                break

            atoms = []
            for site in register:
                new_coordinate = (
                    x_shift + site.coordinate[0], 
                    y_shift + site.coordinate[1]
                )
                parallel_register.add(new_coordinate,site.site_type)

                atoms.append(atom_number)

                atom_number += 1
            
            batch_mapping[(ix,iy)] = atoms
            iy += 1

    return parallel_register,batch_mapping


def parallelize_field(
    field:Field,
    batch_mapping:dict
) -> Field:
    """Generate parallel field from a batch_mapping

    Args:
        field (Field): the field to parallelize
        batch_mapping (dict): the mapping that describes the parallelization

    Returns:
        Field: the new field that works for the parallel program. 
    """
    if field.pattern == None or field.pattern == "uniform":
        return field
    else:
        natoms = sum([len(atom_list) for atom_list in batch_mapping.values()])

        parallel_pattern_series = np.empty(natoms, dtype=object)
        for atom_subset in batch_mapping.values():
            parallel_pattern_series[atom_subset] = field.pattern.series
        
        return Field(field.time_series, Pattern(list(parallel_pattern_series)))
        

def parallelize_hamiltonian(
    hamiltonian: Hamiltonian,
    batch_mapping:dict
) -> DrivingField:
    """Generate the parallel driving fields from a batch_mapping. 

    Args:
        hamiltonian (Hamiltonian): The fields to parallelize
        batch_mapping (dict): the mapping that generates the parallelization

    Returns:
        Hamiltonian: the parallelized driving field. 
    """
    if isinstance(hamiltonian, ShiftingField):
        return ShiftingField(parallelize_field(hamiltonian.magnitude))
    elif isinstance(hamiltonian, DrivingField):
        return DrivingField(
                    amplitude=parallelize_field(hamiltonian.amplitude, batch_mapping),
                    phase=parallelize_field(hamiltonian.phase, batch_mapping),
                    detuning=parallelize_field(hamiltonian.detuning, batch_mapping)
                )
    elif isinstance(hamiltonian, Hamiltonian):
        return Hamiltonian(
                    map(
                        lambda h: parallelize_hamiltonian(h, batch_mapping), 
                        hamiltonian.terms
                    )
                )
    else:
        raise ValueError("expecting Hamiltonian or subtype of Hamiltonian")


def parallelize_ahs(
    ahs:AnalogHamiltonianSimulation,
    qpu: AwsDevice,
    interproblem_distance: Union[float,Decimal]
) -> AnalogHamiltonianSimulation:
    """Generate parallel ahs program. 

    Args:
        ahs (AnalogHamiltonianSimulation): The program to parallelize
        qpu (AwsDevice): The device to run the parallel jobs
        interproblem_distance (float, Decimal): The distance between the programs. 

    Returns:
        AnalogHamiltonianSimulation: The new parallel program ready to run. 
    """
    parallel_register,batch_mapping = generate_parallel_register(ahs.register,qpu,interproblem_distance)

    parallel_program = AnalogHamiltonianSimulation(
        register=parallel_register,
        hamiltonian=parallelize_hamiltonian(ahs.hamiltonian,batch_mapping)
    )
    return parallel_program,batch_mapping


def get_shots_braket_sdk_results(
    results: AnalogHamiltonianSimulationTaskResult,
    batch_mapping:Optional[dict]=None,
    post_select:Optional[bool]=True
)-> np.array:
    """get the shot results from a braket-sdk task results type. 

    Args:
        results (AnalogHamiltonianSimulationTaskResult): 
            The task results to process. 
        batch_mapping (Optional[dict], optional): 
            The parallel mapping generated from some parallelization. Defaults to None.
        post_select (bool, optional): 
        Post select if atom fails to be sorted in the shot results. Defaults to True.

    Returns:
        np.array: The shot results stored as 1,0 in the rows of the array.
        1 is the rydberg state and 0 is the ground state.  
    """
    # collecting QPU Data
    has_defects = lambda x:  np.any(x==0) if post_select else True
    
    if batch_mapping == None:
        all_sequences = [m.post_sequence for m in results.measurements if has_defects(m.pre_sequence)]
    else:
        all_sequences = [m.post_sequence[inds] for m,inds in product(results.measurements,batch_mapping.values()) if has_defects(m.pre_sequence[inds])]
        
    return np.array(all_sequences)


def parallelize_quera_json(
    input_json: dict,
    interproblem_distance:float,
    qpu_width:float,
    qpu_height:float,
    n_site_max:int
) -> Tuple[dict,dict]:
    """Generate a parallel QuEra json program from a single program. 

    Args:
        input_json (dict): The input program to parallelize
        interproblem_distance (float): The distance between parallel problems
        qpu_width (float): The field of view width for the program
        qpu_height (float): The field of view height for the program
        n_site_max (int): Maximum number of sites allowed for a program. 

    Raises:
        NotImplementedError: local detuning currently not supported. 

    Returns:
        Tuple[dict,dict]: first element is the parallelized program as a dict. 
        The second element of the tuple is the batch mapping for post processing. 
    """

    lattice = input_json["lattice"]

    if "local" in input_json["effective_hamiltonian"]["rydberg"]["detuning"]:
        raise NotImplementedError("local detuning not supported in this function")

    else:
        output_json = {}
        output_json.update(input_json)

        sites = lattice["sites"]
        filling = lattice["filling"]
        if len(sites) > 1:
            xmin = min(*[site[0] for site in sites])
            xmax = max(*[site[0] for site in sites])
            ymin = min(*[site[1] for site in sites])
            ymax = max(*[site[1] for site in sites])

            single_problem_width = xmax-xmin
            single_problem_height = ymax-ymin
        else:
            single_problem_width = 0
            single_problem_height = 0


        n_width  = int(qpu_width  // (single_problem_width + interproblem_distance))
        n_height = int(qpu_height //  (single_problem_height + interproblem_distance))

        parallel_sites = []
        parallel_filling = []

        atom_number = 0 #counting number of atoms added
        batch_mapping = {}

        for ix in range(n_width):
            x_shift = np.around(ix * (single_problem_width + interproblem_distance),14)

            for iy in range(n_height):    
                y_shift = np.around(iy * (single_problem_height + interproblem_distance),14)

                # reached the maximum number of batches possible given n_site_max
                if atom_number + len(sites) > n_site_max: break 

                atoms = []
                for site,fill in zip(sites,filling):
                    new_site = [x_shift + site[0], y_shift + site[1]]

                    parallel_sites.append(new_site)
                    parallel_filling.append(fill)

                    atoms.append(atom_number)

                    atom_number += 1

                batch_mapping[f"({ix},{iy})"] = atoms


        output_json["lattice"]["sites"] = parallel_sites
        output_json["lattice"]["filling"] = parallel_filling

    
    return output_json,batch_mapping

def get_shots_quera_results(
    results_json: dict,
    batch_mapping:Optional[dict]=None,
    post_select:Optional[bool]=True
) -> np.array:
    """Get the shots out of a QuEra programming

    Args:
        results_json (dict): _description_
        batch_mapping (dict, optional): 
            The parallel mapping generated from some parallelization. Defaults to None.
        post_select (bool, optional): 
        Post select if atom fails to be sorted in the shot results. Defaults to True.

    Returns:
        np.array: The shot results stored as 1,0 in the rows of the array.
        1 is the rydberg state and 0 is the ground state.
    """
    # collecting QPU Data
    no_defects = lambda bits: np.all(bits==1) if post_select else True
    shots_list = results_json["shot_outputs"]
    
    pre_and_post = [(np.array(m["pre_sequence"]),np.array(m["post_sequence"])) for m in shots_list  if m["shot_status_code"]==200]


    if batch_mapping == None:
        all_sequences = [post for pre,post in pre_and_post  if no_defects(pre)]
    else:
        all_sequences = [post[inds] for (pre,post),inds in product(pre_and_post,batch_mapping.values()) if no_defects(pre[inds])]

    return np.array(all_sequences)



import json
import unittest
import sys,os


from quera_ahs_utils.parallelize import parallelize_quera_json,get_shots_quera_results
import numpy as np


def load_json(filename):
    with open(filename) as file:
        json_value = json.load(file)

    return json_value



class ParallelizeModule(unittest.TestCase):

    def test_1(self) -> None:
        cwd = os.path.dirname(os.path.abspath(__file__))

        input_file = load_json(os.path.join(cwd,"aux_files","test_1.json"))
        output_file = load_json(os.path.join(cwd,"aux_files","parallel_test_1.json"))
        result_file = load_json(os.path.join(cwd,"aux_files","test_1_results_1.json"))

        test_json,batch_mapping = parallelize_quera_json(input_file,1e-5,2.5e-5,2.5e-5,10)

        self.assertDictEqual(test_json,output_file)

        shots = get_shots_quera_results(result_file,batch_mapping=batch_mapping,post_select=True)

        actual_shots = np.array([[1],[0],[1],[0],[1],[0],[0]])
        self.assertTrue(np.all(shots==actual_shots))

        shots = get_shots_quera_results(result_file,batch_mapping=batch_mapping,post_select=False)
        actual_shots = np.array([[1],[0],[1],[0],[1],[0],[0],[0]])

        self.assertTrue(np.all(shots==actual_shots))


        self.assertTrue(np.all(shots==actual_shots))

    def test_2(self) -> None:
        cwd = os.path.dirname(os.path.abspath(__file__))
        result_file = load_json(os.path.join(cwd,"aux_files","test_1_results_1.json"))

        shots = get_shots_quera_results(result_file,batch_mapping=None,post_select=True)
        actual_shots = np.array([[1,0,1,0]])
        self.assertTrue(np.all(shots==actual_shots))

    def test_3(self) -> None:
        cwd = os.path.dirname(os.path.abspath(__file__))
        result_file = load_json(os.path.join(cwd,"aux_files","test_1_results_1.json"))

        shots = get_shots_quera_results(result_file,batch_mapping=None,post_select=False)
        actual_shots = np.array([[1,0,1,0],[1,0,0,0]])
        self.assertTrue(np.all(shots==actual_shots))




from braket.ahs.atom_arrangement import AtomArrangement

import numpy as np
import matplotlib.pyplot as plt
from datetime import datetime 
import os
import sys

from quera_ahs_utils.plotting import plot_avg_density
import unittest

class PlotModule(unittest.TestCase):
    def test_plot_avg_density(self):
        # create register
        register = AtomArrangement()

        # Intentionally non-rectilinear ordering of coordinates
        pos_list = [
            (6.7e-06, 1.34e-05),
            (1.34e-05, 0.0),
            (0.0, 0.0),
            (1.34e-05, 6.7e-06),
            (6.7e-06, 6.7e-06),
            (0.0, 1.34e-05),
            (6.7e-06, 0.0),
            (1.34e-05, 1.34e-05),
            (0.0, 6.7e-06)
        ]

        for pos in pos_list:
            register.add(pos)

        # hardcoded simulation densities
        densities = np.array([0.0075, 0.98, 0.98, 0.01, 0.96, 0.98, 0.0075, 0.9825, 0.01])

        # create matplotlib plot
        fig, ax = plt.subplots()

        # generate plot on custom axis
        plot_avg_density(densities, register, custom_axes=ax)

        # save plot with date and time to file
        now = datetime.now()
        # NOTE: You CANNOT store the date as any format with slashes, this gets misinterpreted as a filepath!
        date_and_time = now.strftime("%m-%d-%Y %H:%M:%S")
        cwd = os.path.dirname(os.path.abspath(__file__))
        filename = os.path.join(cwd,"aux_files",f"{date_and_time}-plot-avg-density-test.png")
        fig.savefig( filename, format="png")



import decimal
import numpy as np

from braket.timings.time_series import TimeSeries
from braket.ahs.driving_field import DrivingField
from braket.ahs.shifting_field import ShiftingField
from braket.ahs.field import Field
from braket.ahs.pattern import Pattern

from numbers import Number
from dataclasses import dataclass
from typing import List, Tuple, Union



__all__ = [
    "rabi_pulse",
    "get_drive",
    "get_shift",
    "constant_time_series",
    "concatenate_time_series",
    "slice_time_series",
    "concatenate_drives",
    "slice_drives",
    "concatenate_shifts",
    "slice_shifts",
    "concatenate_drive_list",
    "concatenate_shift_list",
    
]

def time_series(times: List[float], values: List[float]) -> TimeSeries:
    """create `TimeSeries` object from a list of times and values. 

    Args:
        times (List[float]): :ist of times for TimeSeries
        values (List[float]): List of values for TimeSeries

    Returns:
        TimeSeries: The `TimeSeries` generated from the given times and values. 
    """
    assert len(times) == len(values)
    
    time_series = TimeSeries()
    for time,value in zip(times,values):
        time_series.put(time,value)
    return time_series

def rabi_pulse(
    rabi_pulse_area: float, 
    omega_max: float,
    omega_slew_rate_max: float
) -> Tuple[List[float], List[float]]:
    """Get a time series for Rabi frequency with specified Rabi phase, maximum amplitude
        and maximum slew rate

        Args:
            rabi_pulse_area (float): Total area under the Rabi frequency time series
            omega_max (float): The maximum amplitude 
            omega_slew_rate_max (float): The maximum slew rate

        Returns:
            Tuple[List[float], List[float]]: A tuple containing the time points and values
                of the time series for the time dependent Rabi frequency

        Notes: By Rabi phase, it means the integral of the amplitude of a time-dependent 
            Rabi frequency, \int_0^T\Omega(t)dt, where T is the duration.
    """

    phase_threshold = omega_max**2 / omega_slew_rate_max
    if rabi_pulse_area <= phase_threshold:
        t_ramp = np.sqrt(rabi_pulse_area / omega_slew_rate_max)
        t_plateau = 0
    else:
        t_ramp = omega_max / omega_slew_rate_max
        t_plateau = (rabi_pulse_area / omega_max) - t_ramp
    t_pules = 2 * t_ramp + t_plateau
    time_points = [0, t_ramp, t_ramp + t_plateau, t_pules]
    amplitude_values = [0, t_ramp * omega_slew_rate_max, t_ramp * omega_slew_rate_max, 0]
    
    return time_points, amplitude_values


def get_drive(
    times: List[float], 
    amplitude_values: List[float], 
    detuning_values: List[float], 
    phase_values: List[float]
) -> DrivingField:
    """Get the driving field from a set of time points and values of the fields

        Args:
            times (List[float]): The time points of the driving field
            amplitude_values (List[float]): The values of the amplitude
            detuning_values (List[float]): The values of the detuning
            phase_values (List[float]): The values of the phase

        Returns:
            DrivingField: The driving field obtained
    """

    assert len(times) == len(amplitude_values)
    assert len(times) == len(detuning_values)
    assert len(times) == len(phase_values)
    
    amplitude = TimeSeries()
    detuning = TimeSeries()  
    phase = TimeSeries()    
    
    for t, amplitude_value, detuning_value, phase_value in zip(times, amplitude_values, detuning_values, phase_values):
        amplitude.put(t, amplitude_value)
        detuning.put(t, detuning_value)
        phase.put(t, phase_value) 

    drive = DrivingField(
        amplitude=amplitude, 
        detuning=detuning, 
        phase=phase
    )    
    
    return drive


def get_shift(times: List[float], values: List[float], pattern: List[float]) -> ShiftingField:
    """Get the shifting field from a set of time points, values and pattern

        Args:
            times (List[float]): The time points of the shifting field
            values (List[float]): The values of the shifting field
            pattern (List[float]): The pattern of the shifting field

        Returns:
            ShiftingField: The shifting field obtained
    """    
    assert len(times) == len(values)    
    
    magnitude = TimeSeries()
    for t, v in zip(times, values):
        magnitude.put(t, v)
    shift = ShiftingField(Field(magnitude, Pattern(pattern)))

    return shift


def get_time_series_value(series: Union[TimeSeries,Field], time: float, piecewise_constant: bool = False) -> float:
    """obtain value of a time series at a specific time

    Args:
        time_series (TimeSeries): The `TimeSeries` object to sample from.
        time (float): The time point to sample `times_series` at. 
        piecewise_constant (bool, optional): Interpolate `time_series` assuming piecewise constant, 
            otherwise use piecewise linear. Defaults to False.

    Returns:
        float: the resulting sampled point. 
    """
    if isinstance(series, TimeSeries):
        times = series.times()
        values = series.values()
    elif isinstance(series, Field):
        times = series.time_series.times()
        values = series.time_series.values()

    if piecewise_constant:
        index = max(0, min(len(times) - 1, np.searchsorted(times, time, side="right") - 1))
        return values[index]
    else:
        return np.interp(
            time,
            times,
            values,
            left=values[0],
            right=values[-1]
        )

@dataclass
class DrivingFieldValue:
    amplitude: Number
    detuning: Number
    phase: Number

def get_driving_field_values(driving_field: DrivingField, time: float) -> DrivingFieldValue:
    """Sample the driving feild values at a specific time point.

    Args:
        driving_field (DrivingField): the `DrivingField` object to sample from
        time (float): time to sample `driving_field` at. 

    Returns:
        DrivingFieldValue: An object that contains the values for the different driving fields
    """
    
    return DrivingFieldValue(
        amplitude = get_time_series_value(driving_field.amplitude, time),
        detuning = get_time_series_value(driving_field.detuning, time),
        phase = get_time_series_value(driving_field.phase, time, piecewise_constant=True),
    )


def constant_time_series(other_time_series: TimeSeries, constant: float=0.0) -> TimeSeries:
    """Obtain a constant time series with the same time points as the given time series

        Args:
            other_time_series (TimeSeries): The given time series

        Returns:
            TimeSeries: A constant time series with the same time points as the given time series
    """
    ts = TimeSeries()
    for t in other_time_series.times():
        ts.put(t, constant)
    return ts


def concatenate_time_series(time_series_1: TimeSeries, time_series_2: TimeSeries) -> TimeSeries:
    """Concatenate two time series to a single time series

        Args:
            time_series_1 (TimeSeries): The first time series to be concatenated
            time_series_2 (TimeSeries): The second time series to be concatenated

        Returns:
            TimeSeries: The concatenated time series

    """
    assert time_series_1.values()[-1] == time_series_2.values()[0]
    
    duration_1 = time_series_1.times()[-1] - time_series_1.times()[0]
    
    new_time_series = TimeSeries()
    new_times = time_series_1.times() + [t + duration_1 - time_series_2.times()[0] for t in time_series_2.times()[1:]]
    new_values = time_series_1.values() + time_series_2.values()[1:]
    for t, v in zip(new_times, new_values):
        new_time_series.put(t, v)
    
    return new_time_series


def concatenate_drives(drive_1: DrivingField, drive_2: DrivingField) -> DrivingField:
    """Concatenate two driving fields to a single driving field

        Args:
            drive_1 (DrivingField): The first driving field to be concatenated
            drive_2 (DrivingField): The second driving field to be concatenated

        Returns:
            DrivingField: The concatenated driving field
    """    
    return DrivingField(
        amplitude=concatenate_time_series(drive_1.amplitude.time_series, drive_2.amplitude.time_series),
        detuning=concatenate_time_series(drive_1.detuning.time_series, drive_2.detuning.time_series),
        phase=concatenate_time_series(drive_1.phase.time_series, drive_2.phase.time_series)
    )


def concatenate_shifts(shift_1: ShiftingField, shift_2: ShiftingField) -> ShiftingField:
    """Concatenate two driving fields to a single driving field

        Args:
            shift_1 (ShiftingField): The first shifting field to be concatenated
            shift_2 (ShiftingField): The second shifting field to be concatenated

        Returns:
            ShiftingField: The concatenated shifting field
    """        
    assert shift_1.magnitude.pattern.series == shift_2.magnitude.pattern.series
    
    new_magnitude = concatenate_time_series(shift_1.magnitude.time_series, shift_2.magnitude.time_series)
    return ShiftingField(Field(new_magnitude, shift_1.magnitude.pattern))


def concatenate_drive_list(drive_list: List[DrivingField]) -> DrivingField:
    """Concatenate a list of driving fields to a single driving field

        Args:
            drive_list (List[DrivingField]): The list of driving fields to be concatenated

        Returns:
            DrivingField: The concatenated driving field
    """        
    drive = drive_list[0]
    for dr in drive_list[1:]:
        drive = concatenate_drives(drive, dr)
    return drive    


def concatenate_shift_list(shift_list: List[ShiftingField]) -> ShiftingField:
    """Concatenate a list of shifting fields to a single driving field

        Args:
            shift_list (List[ShiftingField]): The list of shifting fields to be concatenated

        Returns:
            ShiftingField: The concatenated shifting field
    """            
    shift = shift_list[0]
    for sf in shift_list[1:]:
        shift = concatenate_shifts(shift, sf)
    return shift


def slice_time_series(time_series: TimeSeries, first: float, last: float, piecewise_constant=False, min_time_step: float = 0.0):
    """Obtain a sub-section of a TimeSeries between times `first` and `last`

        Args:
            time_series (TimeSeries): The time series to slice
            first (float): Lower bound of the slicing region
            last (float): Upper bound of the slicing region
            piecewise_constant (bool, optional): Flag to use piecewise constant interpolation to get 
                end points of slice, otherwise use piecewise linear interpolation. Defaults to False.
            min_time_step (float): Check if the slice will fall too close to the value the input time series.
                default value is 0.
                
        Returns:
            TimeSeries: The resulting time series after slicing. 
            
        Raises: ValueError
            When the sliced time series will have a time-step smaller than `min_time_step`. 
    """
    times = np.array(time_series.times())    
    values = np.array(time_series.values())
    
    assert first < last
    assert first >= 0
    assert last <= times[-1]
    
    if np.any((times[times > first] - first) < min_time_step):
        raise ValueError("first time step too close to existing time-point.")
    
    if np.any((times[times > last] - last) < min_time_step):
        raise ValueError("first time step too close to existing time-point.")
    
    first_index = np.searchsorted(times,first)
    last_index = np.searchsorted(times,last)

    first_value = get_time_series_value(time_series, first, piecewise_constant=piecewise_constant)
    last_value  = get_time_series_value(time_series, last, piecewise_constant=piecewise_constant)

    new_time_series = TimeSeries()
    
    new_time_series.put(0.0,first_value)
    
    inter_times = times[first_index:last_index]
    inter_values = values[first_index:last_index]
    for time,value in zip(inter_times,inter_values):
        new_time_series.put(time-first,value)
        
    new_time_series.put((last-first),last_value)
    
    return new_time_series

def slice_drive(drive: DrivingField, first: float, last: float, min_time_step: float=0.0) -> DrivingField:
    """Obtain a sub-section of a driving field between times `first` and `last`

        Args:
            drive (DrivingField): The driving field object to be sliced. 
            first (float): Lower bound of the slice. 
            last (float): Upper bound of the slice.

        Returns:
            DrivingField: The resulting driving field after slicing
    """
    return DrivingField(
        amplitude=slice_time_series(drive.amplitude.time_series, first, last, min_time_step=min_time_step),
        detuning=slice_time_series(drive.detuning.time_series, first, last, min_time_step=min_time_step),
        phase=slice_time_series(drive.phase.time_series, first, last, piecewise_constant=True, min_time_step=min_time_step)
    )
    
def slice_shift(shift: ShiftingField, first: float, last: float, min_time_step: float = 0.0) -> ShiftingField:
    """Obtain a sub-section of shifting field begin times `first` and `last`

    Args:
        shift (ShiftingField): The shifting field object to be sliced
        first (float): Lower bound of the slice. 
        last (float): Upper bound of the slice. 

    Returns:
        ShiftingField: The resulting shifting field after slicing. 
    """
    new_time_series = slice_time_series(shift.magnitude.time_series, first, last, min_time_step=min_time_step)
    return ShiftingField(Field(new_time_series, shift.magnitude.pattern))


def adiabatic_drive(
        t_ramp_up: float,
        t_ramp_down: float,
        t_sweep: float,
        Omega_max: float,
        Delta_start: float,
        Delta_end: float
    ) -> DrivingField:
    """Create a basic adiabatic pulse. 

    Args:
        t_ramp_up (float): Time to ramp up rabi drive
        t_ramp_down (float): Time to ramp down rabi drive
        t_sweep (float): Time used to sweep detuning
        Omega_max (float): Maximum value of the rabi drive to use during detuning sweep
        Delta_start (float): Start value for detuning
        Delta_end (float): End value for detuning

    Returns:
        DrivingField: The `DrivingField` object with the resulting adiabatic drive. 
    """
    
    times = [0, t_ramp_up, t_ramp_up + t_sweep, t_ramp_up + t_sweep + t_ramp_down]
    detunings = [Delta_start,Delta_start,Delta_end,Delta_end]
    amplitudes = [0, Omega_max, Omega_max, 0]
    
    return DrivingField(
        amplitude=time_series(times,amplitudes),
        detuning=time_series(times,detunings),
        phase=time_series([0,times[-1]],[0, 0])
    )


def local_detuning_state_prep(register_state: List[int], local_detuning_max: float=150.0e6) -> Tuple[DrivingField, ShiftingField]:
    """Generate the DrivingField and ShiftingField required to use local detuning to prepare the register in a product state. 

    Args:
        register_state (List[int]): The product state you would like to prepare
        local_detuning_max (float, optional): The value of local detuning to apply to atoms which NOT to excit. Defaults to 150.0e6.

    Raises:
        ValueError: register_state must be a list of integers with values 0 or 1. 

    Returns:
        Tuple[DrivingField, ShiftingField]: the driving and shifting fields respectively that contains the pulses for the state preparation. 
    """
    
    pattern = []
    for state in register_state:
        if state == 0:
            pattern.append(1.0)
        elif state == 1:
            pattern.append(0.0)
        else:
            raise ValueError("`register_state` must be a list of integers with values 0 or 1 representing the ground and rydberg states respectively.")
    
    # initial ramp up of local detuning
    ramp_up = time_series([0.0,0.07e-6],[0.0,local_detuning_max])
    shifting_field_start = ShiftingField(
        Field(
            time_series=ramp_up,
            pattern=Pattern(pattern)
        )
    )
    
    driving_field_start = DrivingField(
        amplitude=constant_time_series(ramp_up, 0.0),
        phase=constant_time_series(ramp_up, 0.0),
        detuning=constant_time_series(ramp_up, 0.0)
    )
    
    # resonant pi/2 pulse
    amplitude_max =  14.28e6 # these parameters have a 
    amplitude = time_series(
        times = [0.0, 0.05e-6, 0.11e-6, 0.16e-6],
        values = [0.0,amplitude_max, amplitude_max, 0.0]
    )
    
    driving_field_middle = DrivingField(
        amplitude=amplitude,
        phase=constant_time_series(amplitude, 0.0),
        detuning = constant_time_series(amplitude, 0.0)
    )
    
    shifting_field_middle = ShiftingField(
        Field(
            time_series=constant_time_series(amplitude,local_detuning_max),
            pattern=Pattern(pattern)
        )
    )
    
    # ramp down of local detuning
    
    ramp_down = time_series([0.0,0.07e-6],[local_detuning_max,0.0])
    shifting_field_end = ShiftingField(
        Field(
            time_series=ramp_down,
            pattern=Pattern(pattern)
        )
    )
    
    driving_field_end = DrivingField(
        amplitude=constant_time_series(ramp_down, 0.0),
        phase=constant_time_series(ramp_down, 0.0),
        detuning=constant_time_series(ramp_down, 0.0)
    )
    
    total_shifting_field=concatenate_shift_list([shifting_field_start, shifting_field_middle, shifting_field_end])
    total_driving_field=concatenate_drive_list([driving_field_start, driving_field_middle, driving_field_end])
    
    return total_driving_field, total_shifting_field